package de.tudresden.inf.st.ros2rag.starter;

import com.fasterxml.jackson.databind.ObjectMapper;
import config.Dataconfig;
import de.tudresden.inf.st.ros2rag.starter.ast.*;
import de.tudresden.inf.st.ros2rag.starter.data.DataConfiguration;
import de.tudresden.inf.st.ros2rag.starter.data.DataJoint;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.File;
import java.io.IOException;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

/**
 * Testing Ros2Rag without generating something.
 *
 * @author rschoene - Initial contribution
 */
public class StarterMain {

  private static final Logger logger = LogManager.getLogger(StarterMain.class);
  private MqttUpdater mainHandler;
  private Model model;

  public void run(String[] args) throws IOException, InterruptedException {
    File configFile = new File(args[0]);

    final int[][] safetyZoneCoordinates = {
        {0, 0, 0},
        {-1, 0, 0},
        {1, 0, 0}
    };

    // --- No configuration below this line ---

    ObjectMapper mapper = new ObjectMapper();
    System.out.println("Using config file: " + configFile.getAbsolutePath());
    DataConfiguration config = mapper.readValue(configFile, DataConfiguration.class);

    model = new Model();
    model.MqttSetHost(config.mqttHost);

    ZoneModel zoneModel = new ZoneModel();
    zoneModel.setSize(makePosition(1, 1, 1));

    Zone safetyZone = new Zone();
    for (int[] coordinate : safetyZoneCoordinates) {
      safetyZone.addCoordinate(new Coordinate(
          IntPosition.of(coordinate[0], coordinate[1], coordinate[2])));
    }
    zoneModel.addSafetyZone(safetyZone);
    model.setZoneModel(zoneModel);
    model.MqttWaitUntilReady(2, TimeUnit.SECONDS);

    RobotArm robotArm = new RobotArm();
    model.setRobotArm(robotArm);

    for (DataJoint dataJoint : config.joints) {
      final Joint jointOrEndEffector;
      if (dataJoint.isEndEffector) {
        EndEffector endEffector = new EndEffector();
        robotArm.setEndEffector(endEffector);
        jointOrEndEffector = endEffector;
      } else {
        Joint joint = new Joint();
        robotArm.addJoint(joint);
        jointOrEndEffector = joint;
      }
      jointOrEndEffector.setName(dataJoint.name);
      jointOrEndEffector.setCurrentPosition(makePosition(0, 0, 0));
      robotArm.addDependency1(jointOrEndEffector);
      jointOrEndEffector.connectCurrentPosition(dataJoint.topic);
    }

    robotArm.connectAppropriateSpeed(config.robotConfigTopic, true);

    logStatus("Start", robotArm);
    CountDownLatch exitCondition = new CountDownLatch(1);

    logger.info("To print the current model states, send a message to the topic 'model'.");
    logger.info("To exit the system cleanly, send a message to the topic 'exit', or use Ctrl+C.");

    mainHandler = new MqttUpdater("mainHandler");
    mainHandler.setHost(config.mqttHost);
    mainHandler.waitUntilReady(2, TimeUnit.SECONDS);
    mainHandler.newConnection("exit", bytes -> exitCondition.countDown());
    mainHandler.newConnection("model", bytes -> logStatus(new String(bytes), robotArm));

    sendInitialDataConfig(mainHandler, config.dataConfigTopic);

    Runtime.getRuntime().addShutdownHook(new Thread(this::close));

    exitCondition.await();

    this.close();
  }

  private void sendInitialDataConfig(MqttUpdater mainHandler, String dataConfigTopic) {
    Dataconfig.DataConfig dataConfig = Dataconfig.DataConfig.newBuilder()
        .setEnablePosition(true)
        .setEnableOrientation(false)
        .setEnableTwistAngular(false)
        .setEnableTwistLinear(false)
        .setPublishRate(1000)
        .build();
    mainHandler.publish(dataConfigTopic, dataConfig.toByteArray());
  }

  private void logStatus(String prefix, RobotArm robotArm) {
    StringBuilder sb = new StringBuilder(prefix).append("\n")
        .append("robotArm.isInSafetyZone: ").append(robotArm.isInSafetyZone())
        .append(", robotArm.getAppropriateSpeed = ").append(robotArm.getAppropriateSpeed()).append("\n");
    for (Joint joint : robotArm.getJointList()) {
      sb.append(joint.getName()).append(": ").append(joint.getCurrentPosition()).append("\n");
    }
    sb.append("endEffector ").append(robotArm.getEndEffector().getName()).append(": ")
        .append(robotArm.getEndEffector().getCurrentPosition());
    logger.info(sb.toString());
  }

  private static IntPosition makePosition(int x, int y, int z) {
    return IntPosition.of(x, y, z);
  }

  private void close() {
    logger.info("Exiting ...");
    mainHandler.close();
    model.MqttCloseConnections();
  }

  public static void main(String[] args) throws IOException, InterruptedException {
    new StarterMain().run(args);
  }
}
